#include <Servo.h>
#include <Stepper.h>

/* PINS
 *  
 *  distance sensor: 2, 3
 *  steppers: 4, 5, 6, 7
 *  spine LEDs: 8
 *  roar buzzer: 9
 *  biting servo: 10
 *  tail servo: 11
 *  LDR: A5
 *  eye LEDs: 12
 */


/* https://www.arduino.cc/reference/en/libraries/stepper/
 *  
 *  the driver uses IN1 IN3 IN2 IN4, so the pins are put in that order
 *  
 */
const int stepsPerRevolution = 500; 
Stepper myStepper = Stepper(stepsPerRevolution, 4, 6, 5, 7);
int bodyPos = 0;

/* https://www.arduino.cc/reference/en/libraries/servo/
 *  
 *  sets up a servo to use as jaw and tail and reserves the relevant pins
 *  
 */
Servo biteServo;
static int biteServoPin = 10;

Servo tailServo;
static int tailServoPin = 11;

// reserves the pins for the ultrasound sensor
static int trigPin = 2;
static int echoPin = 3;

// reserves the analog pin for the LDR
static int biteReciever = A5;

// constants for how far the jaw opens when biting
static int openPosition = 0;
static int closedPosition = 20;

// biteTimer is used so the bite can be mostly asynchronous, biteDelay is how often in ms the robot will try to bite again
int biteTimer = -1;
static int biteDelay = 250;

// reserves the pin for the active buzzer to make noise
static int roarPin = 9; 

static int spineLED = 8;
static int spineDelay = 100;
int spineTimer = 0;
bool spineOn = true;

static int eyeLED = 12;
static int eyeDelay = 100;
int eyeTimer = 0;
bool eyeOn = true;

void setup() {
  Serial.begin(9600);
  // setting up pins for ultrasound
  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);
  // setting up pins for the eye and spine LEDS
  pinMode(spineLED, OUTPUT);
  pinMode(eyeLED, OUTPUT);
  // setting up pin for making noise
  pinMode(roarPin, OUTPUT);

  // sets speed in RPM for the stepper, anything above 70 gives inconsistent results but i'm not sure why
  myStepper.setSpeed(60);

  // has the bite servo know to use its pin
  biteServo.attach(biteServoPin);
  tailServo.attach(tailServoPin);
}

void loop() {
  tailServo.write(0);
  delay(2);
  // used for asynchronous running
  int now = millis(); 

  // takes the average of 3 measurements from the ultrasonic sensor and commands the steppers to move the right amount
  float distance = measureDistance(3);
  bodyPos = moveBodyStepper(distance);
  myStepper.step(bodyPos);

  /*
   * this if block checks the LDR to see wether something is in the mouth
   * it also checks wether the robot is ready to bite again
   * if both of these are true, the current time is saved, the mouth is closed and the robot starts roaring
   */
  if (checkBite() and biteTimer == -1) { 
    biteTimer = now;
    closeMouth();
    startRoar();
  }

  /*
   * if the correct amount of time has elapsed, close the mouth, say the robot is ready to bite again and stop making noise
   */
  if (now - biteTimer >= biteDelay) {
    openMouth(); 
    biteTimer = -1;
    endRoar();
  }

  // blinks the spine LEDs every 100ms
  if (now - spineTimer >= spineDelay){
    spineTimer = now;
    spineOn = not(spineOn);
    digitalWrite(spineLED, spineOn);
  }
  // calculates the delay for the eyes and blinks them when necessary
  if (now - eyeTimer >= eyeDelay){
    eyeDelay = calculateEyeDelay(distance);
    eyeTimer = now;
    eyeOn = not(eyeOn);
    digitalWrite(eyeLED, eyeOn);
  }
  
}

int moveBodyStepper(float distance){
  /* written by natalie
   * 
   * this uses a proportional controller to adjust the position of the robot.
   * it checks the output isn't too big so the robot doesn't stop doing everything else
   */
  static float target = 15; //target distance in cm 
  static int gain = 5;
  int error = target - distance;

  // makes sure the error is bounded between -20 and 20
  error = min(20, error);
  error = max(-20, error);
  return error * gain;
}

int moveBodyServo(float distance) {
  /* written by natalie
   * 
   * DEPRECATED
   * 
   * takes in a distance from the sensor in cm and returns a value to move by.
   * this can be used by the servo or the stepper motor
   * 
   */
  float tolerance = 5; //the dinosaur can be 10cm off and not worry about it
  float target = 20;
  int retVal = 0;
  
  if (distance<10.0){
    retVal = 0;
  } 
  if (target<distance){
    retVal = 180;
  }
  if (10.0<distance<target){
    retVal = (180/30) * distance;
  }

  return retVal;
}

bool checkBite() {
  /* written by natalie
   * 
   * checks if the dinosaur needs to bite
   * the threshold may need to be adjusted based on ambient light levels
   * 
   */
   static int threshold = 375;
   int value = analogRead(biteReciever);
   Serial.println(value);
   // change this to return value < threshold
   if (value < threshold) {
    return true;
   } else{
    return false;
   }
}

int closeMouth() {
  /* written by natalie
   * 
   * this closes the mouth to bite down and also records a time, used for asynchronous functions
   * 
   */
  biteServo.write(closedPosition);
  return millis();
}

void openMouth() {
  /* written by natalie
   * 
   * this returns the mouth to its natural position
   * 
   */
  
  biteServo.write(openPosition);
}

float measureDistance(int repetitions) { 
  /* written by natalie
   * 
   * checks the distance sensor 3 times and finds the mean of all the values
   * this should stop any erroneous readings from having too large of an impact
   * it would be possible for more complex error detection, but this would be more computationally intensive
   * 
   */
  float total = 0.0;
  for (int i = 0; i < repetitions; i++){
    total += distanceReading();
  }
  return total/repetitions;
  
}

float distanceReading() {
  /* written by natalie
   * 
   * this function makes the main loop look more clean
   * the number is returned in centimeters
   * it assumes the speed of sound is 340m/s
   * 
   */
  digitalWrite(trigPin, 0);
  delayMicroseconds(5);
  digitalWrite(trigPin, 1);
  delayMicroseconds(10);
  digitalWrite(trigPin, 0);

  long duration = pulseIn(echoPin, HIGH);
  return (duration * 0.034) / 2;
}

void startRoar() {
  /* written by natalie
   * 
   * this function makes the buzzer buzz
   * 
   */ 
  digitalWrite(roarPin, 1);
}

void endRoar() {
   /* written by natalie
   * 
   * this function stops the roar
   * 
   */ 
  digitalWrite(roarPin, 0);
}

int calculateEyeDelay(int distance) {
  /* made by tanaka
   * 
   * this function calculates the delay for the LED blinks
   *  
  */ 
  int out = 0;
  if (distance < 15) {
    out = 300 - (40 * (1/distance));
  }else{
    out = 300;
  }
  
  return 300;
}
